'''Classes to contain and manipulate DNA, RNA, and protein sequences.'''
from ._dna import DNA
from ._dna import RestrictionSite
from ._dna import Primer
from ._peptide import Peptide
from ._rna import RNA
from ._sequence import Feature
